"""
Class: Links

Description:

    This is a plugin/processor that loaded with the Builder class (builder.py)
    It looks for links within markdown files and formatted in the Obsidian/Roam
    style and links them to others within the vault. 

    This is internal links, meaning links within the vault. Extenal links are 
    handled by the markdown renderer. 

Methods:
    __process_matches
    run

Issues:
    - Would like to remove the hard formmated svelte links by adding a config
      file that loads the link format. 
    - One other possible issue is overlaps with other "link" like structures. 
"""

import re

def __process_matches(page, content, matches):
    """
    Parameters:
        content: The content of the markdown page.
        matches: All links within the page.
    """
    url = '<a href="/{}.html">{}</a>'
    for match in matches:
        to_replace = match[0].strip()
        href = match[1].strip()
        content = content.replace(to_replace, url.format(href, href))
    return content

def __get_all_possible_links(content):
    """
    Parameters:
        content: The content of the Markdown page.

    Finds all links while avoid images (!)
    """
    return re.findall("[^$!](\[\[(.*?)\]\])", content)

def run(page):
    content = page.rendered
    matches = __get_all_possible_links(content)
    if len(matches) > 0:
        page.rendered = __process_matches(page, content, matches)
    return page 



import logging
import os
import sys
from pathlib import Path

logger = logging.getLogger(__name__)

def does_vault_path_exist(path):
    path = Path(path) 
    if not path.exists():
        logger.error("ERROR: Cannot find the vault path {}.".format(path))
        return False
    return True 

def is_valid_vault(path):
    theme_path = Path(os.path.join(path, '.theme'))
    if not theme_path.exists():
        logger.error("ERROR: Cannont find theme folder (.theme) in the vault root.")
        return False

    page_path = Path(os.path.join(theme_path, 'page.html'))
    css_path = Path(os.path.join(theme_path, 'page.css'))
    if not page_path.exists() or not css_path.exists():
        logger.error("ERROR: Theme folder (.theme) at vault root must contain the template (page.html) and the stylesheet (page.css).")
        return False
    return True

def prepare_build_path(path):
    path = Path(path)
    if not path.exists():
        path.parent.mkdir(exist_ok=True, parents=True)


"""
Class: Images
pageescription:
page This takes in the profile and rendered content and builds the site fro pageit. It will
   create a menu, move assets and add headers (CSS) as found in the template folder.

Methods:
    __is_image
    __process_matches
    __get_all_possible_images
    __run

Issues:
    - Doesn't support resized images
    - Doesn't support images in subfolders (see moving assets in SiteBuilder)
    - Needs to programmatically detect image types

"""
import re
from pathlib import Path



def __replace_image(content, to_replace, image):
    """
    We need the first check and addition of the "assets/" due to the way Obsidian 
    handles images, leaving some with the full path and others without. 
    """
    if not re.search(r"assets.*", image):
        image = "assets/" + image
    image = '<img src="/{}" alt="{}">'.format(image, image)
    return content.replace(to_replace, image)

def __is_image(image):
    """
    TODO: 
        - Add support for programmatic detection of image types.
    """
    suffix = Path(image).suffix[1:]
    suffixes = ["jpg", "jpeg", "png", "gif", "bmp", "tiff", 
                "svg", "webp", "svg", "avif", "ico", "apng"]
    if suffix.lower() in suffixes:
        return True
    return False

def __process_matches(content, matches):
    for match in matches:
        to_replace = match[0].strip()
        image = match[1].strip()
        if __is_image(image):
            content = __replace_image(content, to_replace, image)
    return content

def __get_all_possible_images(content):
    """
    Get all images in the format ![[ image name here ]]
    """
    return re.findall("(!\[\[(.*?)\]\])", content)

def run(page):
    content = page.rendered
    matches = __get_all_possible_images(content)
    if len(matches) > 0:
        page.rendered = __process_matches(content, matches)
    return page


"""
Class: Page 

Description:

    Used to represent each individual markdown page.

Methods:
    __get_section_title
    __get_markdown_link
    __get_markdown_relative_path
    __get_output_path
    __is_index
    __is_navigation_item
    __is_homepage

Issues:
    - See is_homepage function

"""
import os
from pathlib import Path


class Page:
    def __init__(self, source, section, content, path):
        self.source = source
        self.section = section
        self.content = content
        self.path = path
        self.headers = []
        self.rendered = ""
        self.section_title = self.section.name
        self.filename = Path(self.path).name
        self.markdown_link = self.__get_markdown_link()
        self.markdown_relative_path = self.__get_markdown_relative_path()
        self.output_path = self.__get_output_path()
        self.is_index = self.__is_index()
        self.is_navigation_item = self.__is_navigation_item()
        self.is_homepage = self.__is_homepage()

    def __get_markdown_link(self):
        return self.filename.strip()[:-3]

    def __get_markdown_relative_path(self):
        # print(self.path, self.source)
        # print('test', str(os.path.relpath(self.path, self.source))[:-3])
        return str(os.path.relpath(self.path, self.source))[:-3]

    def __get_output_path(self):
        return self.markdown_relative_path.replace(" ", "_") + ".html"

    def __is_index(self):
        if self.filename.lower() == "index.md":
            return True
        return False

    def __is_navigation_item(self):
        """
        If it's an index.md in one of the folders at the root level then it 
        gets a navigation link.
        """
        if self.is_index:
            path = Path(os.path.relpath(self.path, self.source))
            if len(path.parents) == 2:
                return True
        return False

    def __is_homepage(self):
        """
        If the index.md file as at the root level then it's our homepage.

        Issues: 
            - This is a really bad way of doing this since it checks all files. 
        """
        if self.is_index:
            path = Path(os.path.relpath(self.path, self.source))
            if len(path.parents) == 1:
                return True
        return False

    def display(self):
        """
        Used for debugging
        """
        print("Section:", self.section)
        print("Path:", self.path)
        print("Output path:", self.output_path)
        print("MD link:", self.markdown_link)
        print("MD rel link:", self.markdown_relative_path)
        print("Filename:", self.filename)
        print("Is homepage:", self.is_index)
        print("Is base homepage:", self.is_navigation_item)
        print("Content:", self.content)
        print("Rendered:", self.rendered)



"""
Class: Profile

Description:

    Contains information about the structure of the vault being parsed. 

    Assumptions are made such as: 
        - All folders aside from assets and .obsidian will be rendered
        - All assets, such as images, are stored in the assets folder. 
        - Both content and nested folders may be nested


Methods:
    __get_sections
    __is_valid_file
    __get_asset_paths

Issues:
    - It would be nice to include assets stored in any folder. 

"""
import os
import logging
from pathlib import Path


class SiteProfile:
    def __init__(self, source):
        """
        Parameters:
            source: base folder of the project
        """
        self.source = source
        self.sections = self.__get_sections()
        self.asset_paths = self.__get_assets_paths()

    def __get_sections(self):
        """
        Gets all folders which are not the asset folder. These files should
        contain markdown to be rendered.
        """
        try:
            root = Path(self.source)
            sections = {"root": []}
            for path in Path(root).rglob("*"):
                if not self.__is_valid_path(path):
                    continue
                if os.path.isfile(path):
                    if path.parent in sections:
                        sections[path.parent].append(path)
                    else:
                        sections[path.parent] = [path]
            return sections

        except Exception:
            logging.exception("Error getting sections!")

    def __is_valid_path(self, path):
        """
        Checks if the path is valid. A path is valid if it's:
            1. Not in the assets or .obsidian folders.
            2. Is a markdown file.
        """
        excluded = [
            os.path.join(self.source, exclude) for exclude in ["assets", ".obsidian"]
        ]
        for exclude in excluded:
            if str(path).startswith(str(exclude)):
                return False
        if Path(path).suffix.lower() == ".md":
            return True
        return False

    def __get_assets_paths(self):
        """
        Gets the assets paths as key with the filename name as value.

        Issues: Use os walk to get full tree and scan that way.
        """
        try:
            asset_paths = {}
            path = os.path.join(self.source, "assets")
            for asset_path in Path(path).rglob("*"):
                asset = os.path.relpath(asset_path, path)
                if os.path.isfile(asset_path):
                    asset_paths[str(asset_path)] = asset
            return asset_paths

        except Exception:
            logging.exception("Error getting assets_paths!")


if __name__ == "__main__":
    course_profile = SiteProfile("../course")
    print("\nSections ------------")
    print(course_profile.sections)

    print("\nAssests Paths ------------")
    print(course_profile.asset_paths)



# """
# Main
# """

import logging
import sys
sys.path.append(os.path.dirname(__file__)) # Still need to fix this

from extensions import images, internal_links, sidenotes
from path_check import *
from .site_builder import SiteBuilder
from .site_profile import SiteProfile
from .site_render import SiteRender

logger = logging.getLogger(__name__)

def main():
    if len(sys.argv) != 3:
        print("vault_to_site: incorrect number of arguments.")
        print("Use: python main.py <vault-path> <output-path>")
        quit()

    input_path = sys.argv[1]
    build_path = sys.argv[2]
    if not does_vault_path_exist(input_path):
        quit('')
        
    if not is_valid_vault(input_path):
        quit()

    prepare_build_path(build_path)

    profile = SiteProfile(input_path)
    render = SiteRender(profile, [images, internal_links, sidenotes])
    builder = SiteBuilder(profile, render, build_path)
    builder.copy_assets()
    builder.save_site()
    print('Site rendered')

if __name__ == "__main__":
    main()


"""
Class: Sidenotes

Description:

    Creates sidenotes along the right side. See the global.css for the styling 
    of the aside tag.  

Methods:
    __get_possible_sidenotes
    __process_matches
    run

"""
import re


def __get_possible_sidenotes(content):
    """
    Sidenotes are creating in Obsidian using the pattern ${{ place note here }}
    """
    return re.findall("(\${{(.*?)}})", content)

def __process_matches(content, matches):
    """
    Hacky bit here is closing a p tag and opening again. Since aside tags break the 
    original p tag, if there is one, it'll cause the browser to try and close the one
    that started the paragraph. The closing one then becomes <p></p>, leaving the
    remaining portion of the paragraph unformatted. 
    """
    for match in matches:
        to_replace = match[0].strip()
        tag_title = match[1].strip()
        sidenote = "</p><aside>" + tag_title + "</aside><p>"
        content = content.replace(to_replace, sidenote)
    return content

def run(page):
    content = page.rendered
    matches = __get_possible_sidenotes(content)
    if len(matches) > 0:
        page.rendered = __process_matches(content, matches)
    return page



"""
Class: SiteRender

Description:

   This takes in the profile and rendered content and builds the site from it. It will
   create a menu, move assets and add headers (CSS) as found in the template folder.

Methods:
   copy_assets
   save_site
    __get_header
    __get_page_template
    __get_navigation_menu

Issues:

"""
import os
import logging
import re
import shutil
from pathlib import Path


class SiteBuilder:
    def __init__(self, profile, render, build_path):
        """
        Parameters:
            profile: from SiteProfile, contains project file overview
            render: from SiteRender and contains the rendered individual page content
            build_path: where the final generated site will be saved
        """
        self.profile = profile
        self.render = render
        self.build_path = build_path

    def copy_assets(self):
        """
        Copy all assets from the assets folder, including the primary stylesheet,
        replacing spaces with underscores to avoid issues with the url.
        """

        for asset_source in self.profile.asset_paths:
            asset_dest = os.path.relpath(asset_source, self.profile.source)
            asset_dest = asset_dest.replace(" ", "_")
            asset_dest = Path(os.path.join(self.build_path, asset_dest))
            asset_dest.parent.mkdir(exist_ok=True, parents=True)
            try: 
                shutil.copy(asset_source, asset_dest)
            except Exception: 
                logging.exception("Error copying ")

        try: 
            css_source = os.path.join(self.profile.source, '.theme/page.css')
            css_dest = os.path.join(self.build_path, 'assets/page.css')
            shutil.copy(css_source, css_dest)
        except Exception: 
            logging.exception("Error copying page.css to asssets folder!")

    def save_site(self):
        """
        This uses the page.html template and adds the head information along with the
        menu. It then writes these files out to the build folder.
        """
        header = self.__get_header()
        menu = self.__get_navigation_menu(self.render.pages)
        page_template = self.__get_page_template(os.path.join(self.profile.source, '.theme/page.html'))
        page_template = re.sub("{{\s*header\s*}}", header, page_template)

        for page in self.render.pages:
            path = Path(os.path.join(self.build_path, page.output_path))
            path.parent.mkdir(exist_ok=True, parents=True)
            try: 
                with open(path, "w") as f:
                    html = re.sub("{{\s*menu\s*}}", menu, page_template)
                    html = re.sub("{{\s*body\s*}}", page.rendered, html)
                    f.write(html)
            except Exception: 
                logging.exception("Error opening output file ({}) for writing!".format(page.path))

    def __get_header(self):
        """
        TODO:
            - Need to add these sorts elsewhere instead of hardcoding them.
        """
        return '<link rel="stylesheet" type="text/css" href="/assets/page.css">'

    def __get_page_template(self, filename):
        """
        Parameters:
            - filename: reads in the template to be used.
        TODO:
            - Provide proper error messages
        """
        try:
            with open(filename) as f:
                return f.read()
        except Exception:
            logging.exception("Error reading page.html template")

    def __get_navigation_menu(self, pages):
        """
        Parameters:
            - Takes in the list of rendered pages.

        If the homepage is with a folder at the 'root' level of the Obsidian vault it
        creates a menu item for it.
        """
        menu = []
        menu_item = '<div class="navigation-item"><a href="/{}">{}</a></div>'
        homepage = ''
        for page in pages:
            if page.is_navigation_item:
                menu.append(menu_item.format(page.output_path, page.section_title))
            if page.is_homepage:
                homepage = '<div class="navigation-item homepage"><a href="/{}">Home</a></div>'
                homepage = homepage.format(page.output_path)
        return homepage + "\n".join(reversed(menu))



"""
Class: SiteRender 

Description:

    This class is in charge of rendering the markdown and running through the 
    extensions. Its primary purpose is to create a list of pages (Page class as
    found in page.py) which can be compiled into the site by the SiteRender class. 

Methods:
    __render
    __build_pages
    __read_file
    __render_html

Issues:
    - Probably should clean up some of the naming scheme (output_markdown)
"""

import logging
import markdown as md
from page import Page


class SiteRender:
    def __init__(self, profile, extensions):
        """
        Parameters:
            profile: from SiteProfile, contains project file overview
            extensions: A list of classes that will be run on the markdown (plugins)
        """
        self.profile = profile
        self.extensions = extensions 
        self.pages = self.__render()

    def __render(self):
        """
        Rendering the individual page content
        """
        pages = self.__build_pages()
        for page in pages:
            page.rendered = md.markdown(page.content, extensions=['fenced_code', 'codehilite', 'sane_lists'])
            page = self.__run_extensions(page) 
        return pages

        # for i in range(len(pages)):
        #     pages[i].rendered = md.markdown(pages[i].content, extensions=['fenced_code', 'codehilite', 'sane_lists'])
        #     pages[i] = self.__run_extension(pages[i]) 

    def __build_pages(self):
        """
        Reads from the root markdown folder and builds a dictionary with the
        path as the key and the markdown content as the value.
        """
        pages = []
        for section in self.profile.sections:
            for path in self.profile.sections[section]:
                pages.append(
                    Page(
                        self.profile.source,
                        section=section,
                        content=self.__read_file(path),
                        path=path,
                    )
                )
        return pages

    def __read_file(self, source):
        """
        Parameters:
            source: Path to a markdown file

        Reads in a markdown file and returns it as a single string.
        """
        try:
            with open(source) as f:
                return f.read()
        except Exception:
            logging.exception("Error reading the markdown file.")
            return "[Error processing this file]"

    def __run_extensions(self, page):
        """
        Parameters:
            Page: A single Page object 
        """
        for extension in self.extensions:
            page = extension.run(page)
        return page

